# Copyright 2011-2013 Kwant authors.
#
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

"""This package contains solvers in individual sub-packages.

One of the solvers will be imported as `default`.  Please import other
sub-packages explicitly if you need them.
"""

__all__ = ['default']
from . import default



# Copyright 2011-2016 Kwant authors.
#
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.


import numpy as np
import scipy.linalg as la
from scipy import sparse
from scipy.sparse import csr_matrix as csr
from pytest import raises
from kwant.physics import DiscreteSymmetry
import kwant
from kwant._common import ensure_rng

def test_projectors():
    """Test setting projectors"""
    cons_vals = [-1, -1, 1, 1, 1, 2]
    U = kwant.rmt.circular(6, 'A', rng=8)
    cons = U.T.conj().dot(np.diag(cons_vals)).dot(U)
    # Get the projectors.
    evals, evecs = np.linalg.eigh(cons)
    # Make sure the ordering is correct.
    assert np.allclose(evals-np.array(cons_vals), 0)
    projectors = [np.reshape(evecs[:, :2], (6, 2)),
                  np.reshape(evecs[:, 2:5], (6, 3)),
                  np.reshape(evecs[:, 5], (6, 1))]
    # Ensure that the projectors sum to a unitary and set them.
    assert np.allclose(sum(projector.dot(projector.conj().T) for projector in
                           projectors), np.eye(6))
    # set_symmetry only accepts sparse matrices, so setting np.array
    # projectors should throw an error.
    raises(TypeError, DiscreteSymmetry, projectors=projectors)
    # Make the projectors sparse. Can now set them without trouble.
    projectors = [sparse.coo_matrix(p) for p in projectors]
    DiscreteSymmetry(projectors=projectors)
    # Break one projector - setting them should now throw an error.
    projectors = [p.toarray() for p in projectors]
    projectors[1][2, :] = 0
    assert not np.allclose(sum(projector.dot(projector.conj().T) for projector
                               in projectors), np.eye(6))
    projectors = [sparse.csr_matrix(p) for p in projectors]
    raises(ValueError, DiscreteSymmetry, projectors=projectors)
    # Test setting square projectors.
    p0 = np.eye(20)
    p0[:, ::2] = 0
    p1 = np.eye(20)
    p1[:, 1::2] = 0
    projectors = [sparse.coo_matrix(p0), sparse.coo_matrix(p1)]
    symmetry = DiscreteSymmetry(projectors=projectors)
    # Check that set_symmetry removes zero columns properly.
    p0 = p0[:, 1::2]
    p1 = p1[:, ::2]
    assert np.allclose(symmetry.projectors[0].toarray(), p0)
    assert np.allclose(symmetry.projectors[1].toarray(), p1)

def test_set_discrete_symm():
    n = 20
    # All discrete symmetries
    sym = 'CII'
    t_mat = np.array(kwant.rmt.h_t_matrix[sym])
    t_mat = np.kron(np.identity(n // len(t_mat)), t_mat)
    p_mat = np.array(kwant.rmt.h_p_matrix[sym])
    p_mat = np.kron(np.identity(n // len(p_mat)), p_mat)
    c_mat = p_mat.dot(t_mat.conj())

    # Set one symmetry at a time.
    symms = [(t_mat, None, None), (None, p_mat, None), (None, None, c_mat)]
    for symm in symms:
        T, P, C = symm
        # Try to set the symmetries as numpy arrays. An error should be thrown.
        raises(TypeError, DiscreteSymmetry, time_reversal=T, particle_hole=P,
               chiral=C)

        s_symm = (sparse.coo_matrix(op) if op is not None else None
                  for op in symm)
        T, P, C = s_symm
        DiscreteSymmetry(time_reversal=T, particle_hole=P, chiral=C)
        # Break the symmetry operator - an error should be thrown.
        b_symm = (sparse.csr_matrix(kwant.rmt.circular(op.shape[0], 'A',
                                                       rng=2))
                  if op is not None else None for op in symm)
        T, P, C = b_symm
        raises(ValueError, DiscreteSymmetry, time_reversal=T, particle_hole=P,
               chiral=C)

    # Set all three symmetries
    t_mat = sparse.csr_matrix(t_mat)
    p_mat = sparse.csr_matrix(p_mat)
    c_mat = sparse.csr_matrix(c_mat)
    DiscreteSymmetry(time_reversal=t_mat, particle_hole=p_mat, chiral=c_mat)

    # Break chiral symmetry - replace it with identity,
    # such that the product of symmetries is incorrect,
    # and an error should the thrown.
    raises(ValueError, DiscreteSymmetry, time_reversal=t_mat,
           particle_hole=p_mat, chiral=sparse.eye(n))
    raises(ValueError, DiscreteSymmetry, time_reversal=2*t_mat)
    raises(ValueError, DiscreteSymmetry, chiral=2*t_mat)

    # Check that with two symmetries specified, the third one is computed.
    symm = DiscreteSymmetry(time_reversal=t_mat, particle_hole=p_mat)
    assert np.allclose(symm.chiral.toarray(), c_mat.toarray())
    DiscreteSymmetry(time_reversal=t_mat, chiral=c_mat)
    assert np.allclose(symm.particle_hole.toarray(), p_mat.toarray())
    DiscreteSymmetry(particle_hole=p_mat, chiral=c_mat)
    assert np.allclose(symm.time_reversal.toarray(), t_mat.toarray())


def test_projectors_and_symmetry():
    # Consider a Hamiltonian with three blocks: the first one has a discrete
    # symmetry and the other two are related by the same discrete symmetry (but
    # do not possess it themselves).  Further, the first block is twice the
    # size of the latter two. For this situation and if the symmetry is chiral
    # or time reversal symmetry, the full symmetry operators are:
    n = 20
    c_mat1 = np.kron(np.identity(n), np.diag([1, -1]))
    c_mat2 = np.kron(np.identity(n // 2), np.diag([1, -1]))
    sx = np.array([[0,1],[1,0]])
    C_mat = la.block_diag(c_mat1, np.kron(sx, c_mat2))
    assert np.allclose(C_mat.dot(C_mat), np.eye(C_mat.shape[0]))
    t_mat = np.array(kwant.rmt.h_t_matrix['CII'])
    t_mat1 = np.kron(np.identity(2*n // len(t_mat)), t_mat)
    t_mat2 = np.kron(np.identity(n // len(t_mat)), t_mat)
    T_mat = la.block_diag(t_mat1, np.kron(sx, t_mat2))
    assert np.allclose(T_mat.dot(T_mat.conj()), -np.eye(T_mat.shape[0]))
    # The projectors that block diagonalize the Hamiltonian are
    I = np.eye(4*n)
    projectors = [I[:, :2*n], I[:, 2*n:3*n], I[:, 3*n:]]
    projectors = [sparse.csr_matrix(p) for p in projectors]
    # The symmetry and projectors are in canonical form and they commute,
    # so declaring should not throw an error or raise a warning.
    U = sparse.csr_matrix(kwant.rmt.circular(4*n, 'A', rng=234))

    for (C, T) in [(sparse.csr_matrix(C_mat), None),
                   (None, sparse.csr_matrix(T_mat))]:
        DiscreteSymmetry(projectors=projectors, chiral=C, time_reversal=T)
        # Transform symmetries with a random unitary, such that they are
        # no longer in canonical form, but leave projectors as is. An error
        # should be thrown.
        if C is not None: C = U.T.conj().dot(C).dot(U)
        if T is not None: T = U.T.conj().dot(T).dot(U.conj())
        raises(ValueError, DiscreteSymmetry, projectors=projectors, chiral=C,
               time_reversal=T)


def test_validate():
    csr = sparse.csr_matrix
    sym = DiscreteSymmetry(projectors=[csr(np.array([[1], [0]])),
                                       csr(np.array([[0], [1]]))])
    assert sym.validate(csr(np.array([[0], [1]]))) == ['Conservation law']
    assert sym.validate(np.array([[1], [0]])) == []
    assert sym.validate(np.eye(2)) == []
    assert sym.validate(1 - np.eye(2)) == ['Conservation law']

    sym = DiscreteSymmetry(particle_hole=sparse.identity(2))
    assert sym.validate(1j * sparse.identity(2)) == []
    assert sym.validate(sparse.identity(2)) == ['Particle-hole']

    sym = DiscreteSymmetry(time_reversal=sparse.identity(2))
    assert sym.validate(sparse.identity(2)) == []
    assert sym.validate(1j * sparse.identity(2)) == ['Time reversal']

    sym = DiscreteSymmetry(chiral=csr(np.diag((1, -1))))
    assert sym.validate(np.eye(2)) == ['Chiral']
    assert sym.validate(1 - np.eye(2)) == []


def random_onsite_hop(n, rng=0):
    rng = ensure_rng(rng)
    onsite = rng.randn(n, n) + 1j * rng.randn(n, n)
    onsite = onsite + onsite.T.conj()
    hop = rng.rand(n, n) + 1j * rng.rand(n, n)
    return onsite, hop


def test_validate_commutator():
    symm_class = ['AI', 'AII', 'D', 'C', 'AIII', 'BDI']
    sym_dict = {'AI': ['Time reversal'],
                'AII': ['Time reversal'],
                'D': ['Particle-hole'],
                'C': ['Particle-hole'],
                'AIII': ['Chiral'],
                'BDI': ['Time reversal', 'Particle-hole', 'Chiral']}
    n = 10
    rng = 10
    for sym in symm_class:
        # Random matrix in symmetry class
        h = kwant.rmt.gaussian(n, sym, rng=rng)
        if kwant.rmt.p(sym):
            p_mat = np.array(kwant.rmt.h_p_matrix[sym])
            p_mat = csr(np.kron(np.identity(n // len(p_mat)), p_mat))
        else:
            p_mat = None
        if kwant.rmt.t(sym):
            t_mat = np.array(kwant.rmt.h_t_matrix[sym])
            t_mat = csr(np.kron(np.identity(n // len(t_mat)), t_mat))
        else:
            t_mat = None
        if kwant.rmt.c(sym):
            c_mat = csr(np.kron(np.identity(n // 2), np.diag([1, -1])))
        else:
            c_mat = None
        disc_symm = DiscreteSymmetry(particle_hole=p_mat,
                                     time_reversal=t_mat,
                                     chiral=c_mat)
        assert disc_symm.validate(h) == []
        a = random_onsite_hop(n, rng=rng)[1]
        for symmetry in disc_symm.validate(a):
            assert symmetry in sym_dict[sym]



# Copyright 2011-2018 Kwant authors.
#
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.


__all__ = ['builder_to_model', 'model_to_builder', 'find_builder_symmetries']

import itertools as it
from collections import OrderedDict, defaultdict

import numpy as np
import tinyarray as ta
import scipy.linalg as la

try:
    import sympy
    import qsymm
    from qsymm.model import Model, BlochModel, BlochCoeff
    from qsymm.groups import PointGroupElement, ContinuousGroupGenerator
    from qsymm.symmetry_finder import bravais_point_group
    from qsymm.linalg import allclose
    from qsymm.hamiltonian_generator import hamiltonian_from_family
except ImportError as error:
    msg = ("'kwant.qsymm' is not available because one or more of its "
           "dependencies is not installed.")
    raise ImportError(msg) from error

from kwant import lattice, builder
from kwant._common import get_parameters


def builder_to_model(syst, momenta=None, real_space=True,
                     params=None):
    """Make a qsymm.BlochModel out of a `~kwant.builder.Builder`.

    Parameters
    ----------
    syst : `~kwant.builder.Builder`
        May have translational symmetries.
    momenta : list of strings or None
        Names of momentum variables. If None, 'k_x', 'k_y', ... is used.
    real_space : bool (default True)
        If False, use the unit cell convention for Bloch basis, the
        exponential has the difference in the unit cell coordinates and
        k is expressed in the reciprocal lattice basis. This is consistent
        with `kwant.wraparound`.
        If True, the difference in the real space coordinates is used
        and k is given in an absolute basis.
        Only the default choice guarantees that qsymm is able to find
        nonsymmorphic symmetries.
    params : dict, optional
        Dictionary of parameter names and their values; used when
        evaluating the Hamiltonian matrix elements.

    Returns
    -------
    model : qsymm.BlochModel
        Model representing the tight-binding Hamiltonian.

    Notes
    -----
    The sites in the the builder are in lexicographical order, i.e. ordered
    first by their family and then by their tag. This is the same ordering that
    is used in finalized kwant systems.
    """
    def term_to_model(d, par, matrix):
        if allclose(matrix, 0):
            result = BlochModel({}, shape=matrix.shape, format=np.ndarray)
        else:
            result = BlochModel({BlochCoeff(d, qsymm.sympify(par)): matrix},
                                momenta=momenta)
        return result

    def hopping_to_model(hop, value, proj, params):
        site1, site2 = hop
        if real_space:
            d = proj @ np.array(site2.pos - site1.pos)
        else:
            # site in the FD
            d = np.array(syst.symmetry.which(site2))

        slice1, slice2 = slices[to_fd(site1)], slices[to_fd(site2)]
        if callable(value):
            return sum(term_to_model(d, par, set_block(slice1, slice2, val))
                       for par, val in function_to_terms(hop, value, params))
        else:
            matrix = set_block(slice1, slice2, value)
            return term_to_model(d, '1', matrix)

    def onsite_to_model(site, value, params):
        d = np.zeros((dim, ))
        slice1 = slices[to_fd(site)]
        if callable(value):
            return sum(term_to_model(d, par, set_block(slice1, slice1, val))
                       for par, val in function_to_terms(site, value, params))
        else:
            return term_to_model(d, '1', set_block(slice1, slice1, value))

    def function_to_terms(site_or_hop, value, fixed_params):
        assert callable(value)
        parameters = get_parameters(value)
        # remove site or site1, site2 parameters
        if isinstance(site_or_hop, builder.Site):
            parameters = parameters[1:]
            site_or_hop = (site_or_hop,)
        else:
            parameters = parameters[2:]
        free_parameters = (par for par in parameters
                           if par not in fixed_params.keys())
        # first set all free parameters to 0
        args = ((fixed_params[par] if par in fixed_params.keys() else 0)
                for par in parameters)
        h_0 = value(*site_or_hop, *args)
        # set one of the free parameters to 1 at a time, the rest 0
        terms = []
        for p in free_parameters:
            args = ((fixed_params[par] if par in fixed_params.keys() else
                     (1 if par == p else 0)) for par in parameters)
            terms.append((p, value(*site_or_hop, *args) - h_0))
        return terms + [('1', h_0)]

    def orbital_slices(syst):
        orbital_slices = {}
        start_orb = 0

        for site in sorted(syst.sites()):
            n = site.family.norbs
            if n is None:
                raise ValueError('norbs must be provided for every lattice.')
            orbital_slices[site] = slice(start_orb, start_orb + n)
            start_orb += n
        return orbital_slices, start_orb

    def set_block(slice1, slice2, val):
        matrix = np.zeros((N, N), dtype=complex)
        matrix[slice1, slice2] = val
        return matrix

    if params is None:
        params = dict()

    periods = np.array(syst.symmetry.periods)
    dim = len(periods)
    to_fd = syst.symmetry.to_fd
    if momenta is None:
        momenta = ['k_x', 'k_y', 'k_z'][:dim]
    # If the system is higher dimensional than the number of translation
    # vectors, we need to project onto the subspace spanned by the
    # translation vectors.
    if dim == 0:
        proj = np.empty((0, len(list(syst.sites())[0].pos)))
    elif dim < len(list(syst.sites())[0].pos):
        proj, r = la.qr(np.array(periods).T, mode='economic')
        sign = np.diag(np.diag(np.sign(r)))
        proj = sign @ proj.T
    else:
        proj = np.eye(dim)

    slices, N = orbital_slices(syst)

    one_way_hoppings = [hopping_to_model(hop, value, proj, params)
                        for hop, value in syst.hopping_value_pairs()]
    other_way_hoppings = [term.T().conj() for term in one_way_hoppings]
    hoppings = one_way_hoppings + other_way_hoppings

    onsites = [onsite_to_model(site, value, params)
               for site, value in syst.site_value_pairs()]

    result = sum(onsites) + sum(hoppings)

    return result


def model_to_builder(model, norbs, lat_vecs, atom_coords, *, coeffs=None):
    """Make a `~kwant.builder.Builder` out of qsymm.Models or qsymm.BlochModels.

    Parameters
    ----------
    model : qsymm.Model, qsymm.BlochModel, or an iterable thereof
        The Hamiltonian (or terms of the Hamiltonian) to convert to a
        Builder.
    norbs : OrderedDict or sequence of pairs
        Maps sites to the number of orbitals per site in a unit cell.
    lat_vecs : list of arrays
        Lattice vectors of the underlying tight binding lattice.
    atom_coords : list of arrays
        Positions of the sites (or atoms) within a unit cell.
        The ordering of the atoms is the same as in norbs.
    coeffs : list of sympy.Symbol, default None.
        Constant prefactors for the individual terms in model, if model
        is a list of multiple objects. If model is a single Model or BlochModel
        object, this argument is ignored. By default assigns the coefficient
        c_n to element model[n].

    Returns
    -------
    syst : `~kwant.builder.Builder`
        The unfinalized Kwant system representing the qsymm Model(s).

    Notes
    -----
    Onsite terms that are not provided in the input model are set
    to zero by default.

    The input model(s) representing the tight binding Hamiltonian in
    Bloch form should follow the convention where the difference in the real
    space atomic positions appear in the Bloch factors.
    """

    def make_int(R):
        # If close to an integer array convert to integer tinyarray, else
        # return None
        R_int = ta.array(np.round(R), int)
        if qsymm.linalg.allclose(R, R_int):
            return R_int
        else:
            return None

    def term_onsite(onsites_dict, hopping_dict, hop_mat, atoms,
                    sublattices, coords_dict):
        """Find the Kwant onsites and hoppings in a qsymm.BlochModel term
        that has no lattice translation in the Bloch factor.
        """
        for atom1, atom2 in it.product(atoms, atoms):
            # Subblock within the same sublattice is onsite
            hop = hop_mat[ranges[atom1], ranges[atom2]]
            if sublattices[atom1] == sublattices[atom2]:
                onsites_dict[atom1] += Model({coeff: hop}, momenta=momenta)
            # Blocks between sublattices are hoppings between sublattices
            # at the same position.
            # Only include nonzero hoppings
            elif not allclose(hop, 0):
                if not allclose(np.array(coords_dict[atom1]),
                                np.array(coords_dict[atom2])):
                    raise ValueError(
                        "Position of sites not compatible with qsymm model.")
                lat_basis = np.array(zer)
                hop = Model({coeff: hop}, momenta=momenta)
                hop_dir = builder.HoppingKind(-lat_basis, sublattices[atom1],
                                              sublattices[atom2])
                hopping_dict[hop_dir] += hop
        return onsites_dict, hopping_dict

    def term_hopping(hopping_dict, hop_mat, atoms,
                     sublattices, coords_dict):
        """Find Kwant hoppings in a qsymm.BlochModel term that has a lattice
        translation in the Bloch factor.
        """
        # Iterate over combinations of atoms, set hoppings between each
        for atom1, atom2 in it.product(atoms, atoms):
            # Take the block from atom1 to atom2
            hop = hop_mat[ranges[atom1], ranges[atom2]]
            # Only include nonzero hoppings
            if allclose(hop, 0):
                continue
            # Adjust hopping vector to Bloch form basis
            r_lattice = (
                r_vec
                + np.array(coords_dict[atom1])
                - np.array(coords_dict[atom2])
            )
            # Bring vector to basis of lattice vectors
            lat_basis = np.linalg.solve(np.vstack(lat_vecs).T, r_lattice)
            lat_basis = make_int(lat_basis)
            # Should only have hoppings that are integer multiples of
            # lattice vectors
            if lat_basis is not None:
                hop_dir = builder.HoppingKind(-lat_basis,
                                              sublattices[atom1],
                                              sublattices[atom2])
                # Set the hopping as the matrix times the hopping amplitude
                hopping_dict[hop_dir] += Model({coeff: hop}, momenta=momenta)
            else:
                raise RuntimeError('A nonzero hopping not matching a '
                                   'lattice vector was found.')
        return hopping_dict

    # Disambiguate single model instances from iterables thereof. Because
    # Model is itself iterable (subclasses dict) this is a bit cumbersome.
    if isinstance(model, Model):
        # BlochModel can't yet handle getting a Blochmodel as input
        if not isinstance(model, BlochModel):
            model = BlochModel(model)
    else:
        model = BlochModel(hamiltonian_from_family(
            model, coeffs=coeffs, nsimplify=False, tosympy=False))


    # 'momentum' and 'zer' are used in the closures defined above, so don't
    # move these declarations down.
    momenta = model.momenta
    if len(momenta) != len(lat_vecs):
        raise ValueError("Dimension of the lattice and number of "
                         "momenta do not match.")
    zer = [0] * len(momenta)


    # Subblocks of the Hamiltonian for different atoms.
    N = 0
    if not any([isinstance(norbs, OrderedDict), isinstance(norbs, list),
                isinstance(norbs, tuple)]):
        raise ValueError('norbs must be OrderedDict, tuple, or list.')
    else:
        norbs = OrderedDict(norbs)
    ranges = dict()
    for a, n in norbs.items():
        ranges[a] = slice(N, N + n)
        N += n

    # Extract atoms and number of orbitals per atom,
    # store the position of each atom
    atoms, orbs = zip(*norbs.items())
    coords_dict = dict(zip(atoms, atom_coords))

    # Make the kwant lattice
    lat = lattice.general(lat_vecs, atom_coords, norbs=orbs)
    # Store sublattices by name
    sublattices = dict(zip(atoms, lat.sublattices))

    # Keep track of the hoppings and onsites by storing those
    # which have already been set.
    hopping_dict = defaultdict(lambda: 0)
    onsites_dict = defaultdict(lambda: 0)

    # Iterate over all terms in the model.
    for key, hop_mat in model.items():
        # Determine whether this term is an onsite or a hopping, extract
        # overall symbolic coefficient if any, extract the exponential
        # part describing the hopping if present.
        r_vec, coeff = key
        # Onsite term; modifies onsites_dict and hopping_dict in-place
        if allclose(r_vec, 0):
            term_onsite(
                onsites_dict, hopping_dict, hop_mat,
                atoms, sublattices, coords_dict)
        # Hopping term; modifies hopping_dict in-place
        else:
            term_hopping(hopping_dict, hop_mat, atoms,
                         sublattices, coords_dict)

    # If some onsite terms are not set, we set them to zero.
    for atom in atoms:
        if atom not in onsites_dict:
            onsites_dict[atom] = Model(
                {sympy.numbers.One(): np.zeros((norbs[atom], norbs[atom]))},
                momenta=momenta)

    # Make the Kwant system, and set all onsites and hoppings.

    sym = lattice.TranslationalSymmetry(*lat_vecs)
    syst = builder.Builder(sym)

    # Iterate over all onsites and set them
    for atom, onsite in onsites_dict.items():
        syst[sublattices[atom](*zer)] = onsite.lambdify(onsite=True)

    # Finally, iterate over all the hoppings and set them
    for direction, hopping in hopping_dict.items():
        syst[direction] = hopping.lambdify(hopping=True)

    return syst


# This may be useful in the future, so we'll keep it as internal for now,
# and can make it part of the API in the future if we wish.
def _get_builder_symmetries(builder):
    """Extract the declared symmetries of a Kwant builder.

    Parameters
    ----------
    builder : `~kwant.builder.Builder`

    Returns
    -------
    builder_symmetries : dict
        Dictionary of the discrete symmetries that the builder has.
        The symmetries can be particle-hole, time-reversal or chiral,
        which are returned as qsymm.PointGroupElements, or
        a conservation law, which is returned as a
        qsymm.ContinuousGroupGenerators.
    """

    dim = len(np.array(builder.symmetry.periods))
    symmetry_names = ['time_reversal', 'particle_hole', 'chiral',
                      'conservation_law']
    builder_symmetries = {name: getattr(builder, name)
                          for name in symmetry_names
                          if getattr(builder, name) is not None}
    for name, symmetry in builder_symmetries.items():
        if name == 'time_reversal':
            builder_symmetries[name] = PointGroupElement(np.eye(dim),
                                                         True, False, symmetry)
        elif name == 'particle_hole':
            builder_symmetries[name] = PointGroupElement(np.eye(dim),
                                                         True, True, symmetry)
        elif name == 'chiral':
            builder_symmetries[name] = PointGroupElement(np.eye(dim),
                                                         False, True, symmetry)
        elif name == 'conservation_law':
            builder_symmetries[name] = ContinuousGroupGenerator(R=None,
                                                                U=symmetry)
        else:
            raise ValueError("Invalid symmetry name.")
    return builder_symmetries


def find_builder_symmetries(builder, momenta=None, params=None,
                            spatial_symmetries=True, prettify=True,
                            sparse=None):
    """Finds the symmetries of a Kwant system using qsymm.

    Parameters
    ----------
    builder : `~kwant.builder.Builder`
    momenta : list of strings or None
        Names of momentum variables, if None 'k_x', 'k_y', ... is used.
    params : dict, optional
        Dictionary of parameter names and their values; used when
        evaluating the Hamiltonian matrix elements.
    spatial_symmetries : bool (default True)
        If True, search for all symmetries.
        If False, only searches for the symmetries that are declarable in
        `~kwant.builder.Builder` objects, i.e. time-reversal symmetry,
        particle-hole symmetry, chiral symmetry, or conservation laws.
        This can save computation time.
    prettify : bool (default True)
        Whether to carry out sparsification of the continuous symmetry
        generators, in general an arbitrary linear combination of the
        symmetry generators is returned.
    sparse : bool, or None (default None)
        Whether to use sparse linear algebra in the calculation.
        Can give large performance gain in large systems.
        If None, uses sparse or dense computation depending on
        the size of the Hamiltonian.


    Returns
    -------
    symmetries : list of qsymm.PointGroupElements and/or qsymm.ContinuousGroupElement
        The symmetries of the Kwant system.
    """

    if params is None:
        params = dict()

    ham = builder_to_model(builder, momenta=momenta,
                           real_space=spatial_symmetries, params=params)

    # Use dense or sparse computation depending on Hamiltonian size
    if sparse is None:
        sparse = list(ham.values())[0].shape[0] > 20

    dim = len(np.array(builder.symmetry.periods))

    if spatial_symmetries:
        candidates = bravais_point_group(builder.symmetry.periods, tr=True,
                                         ph=True, generators=False,
                                         verbose=False)
    else:
        candidates = [
            qsymm.PointGroupElement(np.eye(dim), True, False, None),  # T
            qsymm.PointGroupElement(np.eye(dim), True, True, None),   # P
            qsymm.PointGroupElement(np.eye(dim), False, True, None)]  # C
    sg, cg = qsymm.symmetries(ham, candidates, prettify=prettify,
                              continuous_rotations=False,
                              sparse_linalg=sparse)
    return list(sg) + list(cg)



# Copyright 2011-2013 Kwant authors.
#
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

from libc.stdio cimport FILE

cdef extern from "scotch.h":
    ctypedef struct SCOTCH_Arch:
        pass

    ctypedef struct SCOTCH_Geom:
        pass

    ctypedef struct SCOTCH_Graph:
        double dummy[12]

    ctypedef struct SCOTCH_Mesh:
        pass

    ctypedef struct SCOTCH_Mapping:
        pass

    ctypedef struct SCOTCH_Ordering:
        pass

    ctypedef struct SCOTCH_Strat:
        pass

    ctypedef int SCOTCH_Idx

    ctypedef int SCOTCH_Num

    void SCOTCH_errorProg(char *)
    void SCOTCH_errorPrint(char *, ...)
    void SCOTCH_errorPrintW(char *, ...)

    int SCOTCH_archInit(SCOTCH_Arch *)
    void SCOTCH_archExit(SCOTCH_Arch *)
    int SCOTCH_archLoad(SCOTCH_Arch *, FILE *)
    int SCOTCH_archSave(SCOTCH_Arch *, FILE *)
    int SCOTCH_archBuild(SCOTCH_Arch *, SCOTCH_Graph *, SCOTCH_Num, SCOTCH_Num *, SCOTCH_Strat *)
    char *SCOTCH_archName(SCOTCH_Arch *)
    SCOTCH_Num SCOTCH_archSize(SCOTCH_Arch *)
    int SCOTCH_archCmplt(SCOTCH_Arch *, SCOTCH_Num)
    int SCOTCH_archCmpltw(SCOTCH_Arch *, SCOTCH_Num, SCOTCH_Num *)
    int SCOTCH_archHcub(SCOTCH_Arch *, SCOTCH_Num)
    int SCOTCH_archMesh2(SCOTCH_Arch *, SCOTCH_Num, SCOTCH_Num)
    int SCOTCH_archMesh3(SCOTCH_Arch *, SCOTCH_Num, SCOTCH_Num, SCOTCH_Num)
    int SCOTCH_archTleaf(SCOTCH_Arch *, SCOTCH_Num, SCOTCH_Num *, SCOTCH_Num *)
    int SCOTCH_archTorus2(SCOTCH_Arch *, SCOTCH_Num, SCOTCH_Num)
    int SCOTCH_archTorus3(SCOTCH_Arch *, SCOTCH_Num, SCOTCH_Num, SCOTCH_Num)
    int SCOTCH_archVcmplt(SCOTCH_Arch *)
    int SCOTCH_archVhcub(SCOTCH_Arch *)

    int SCOTCH_geomInit(SCOTCH_Geom *)
    void SCOTCH_geomExit(SCOTCH_Geom *)
    void SCOTCH_geomData(SCOTCH_Geom *, SCOTCH_Num *, double **)

    int SCOTCH_graphInit(SCOTCH_Graph *)
    void SCOTCH_graphExit(SCOTCH_Graph *)
    void SCOTCH_graphFree(SCOTCH_Graph *)
    int SCOTCH_graphLoad(SCOTCH_Graph *, FILE *, SCOTCH_Num, SCOTCH_Num)
    int SCOTCH_graphSave(SCOTCH_Graph *, FILE *)
    int SCOTCH_graphBuild(SCOTCH_Graph *, SCOTCH_Num, SCOTCH_Num, SCOTCH_Num *, SCOTCH_Num *, SCOTCH_Num *, SCOTCH_Num *, SCOTCH_Num, SCOTCH_Num *, SCOTCH_Num *)
    SCOTCH_Num SCOTCH_graphBase(SCOTCH_Graph *, SCOTCH_Num baseval)
    int SCOTCH_graphCheck(SCOTCH_Graph *)
    void SCOTCH_graphSize(SCOTCH_Graph *, SCOTCH_Num *, SCOTCH_Num *)
    void SCOTCH_graphData(SCOTCH_Graph *, SCOTCH_Num *, SCOTCH_Num *, SCOTCH_Num **, SCOTCH_Num **, SCOTCH_Num **, SCOTCH_Num **, SCOTCH_Num *, SCOTCH_Num **, SCOTCH_Num **)
    void SCOTCH_graphStat(SCOTCH_Graph *, SCOTCH_Num *, SCOTCH_Num *, SCOTCH_Num *, double *, double *, SCOTCH_Num *, SCOTCH_Num *, double *, double *, SCOTCH_Num *, SCOTCH_Num *, SCOTCH_Num *, double *, double *)
    int SCOTCH_graphGeomLoadChac(SCOTCH_Graph *, SCOTCH_Geom *, FILE *, FILE *, char *)
    int SCOTCH_graphGeomLoadHabo(SCOTCH_Graph *, SCOTCH_Geom *, FILE *, FILE *, char *)
    int SCOTCH_graphGeomLoadMmkt(SCOTCH_Graph *, SCOTCH_Geom *, FILE *, FILE *, char *)
    int SCOTCH_graphGeomLoadScot(SCOTCH_Graph *, SCOTCH_Geom *, FILE *, FILE *, char *)
    int SCOTCH_graphGeomSaveChac(SCOTCH_Graph *, SCOTCH_Geom *, FILE *, FILE *, char *)
    int SCOTCH_graphGeomSaveMmkt(SCOTCH_Graph *, SCOTCH_Geom *, FILE *, FILE *, char *)
    int SCOTCH_graphGeomSaveScot(SCOTCH_Graph *, SCOTCH_Geom *, FILE *, FILE *, char *)

    int SCOTCH_graphMapInit(SCOTCH_Graph *, SCOTCH_Mapping *, SCOTCH_Arch *, SCOTCH_Num *)
    void SCOTCH_graphMapExit(SCOTCH_Graph *, SCOTCH_Mapping *)
    int SCOTCH_graphMapLoad(SCOTCH_Graph *, SCOTCH_Mapping *, FILE *)
    int SCOTCH_graphMapSave(SCOTCH_Graph *, SCOTCH_Mapping *, FILE *)
    int SCOTCH_graphMapView(SCOTCH_Graph *, SCOTCH_Mapping *, FILE *)
    int SCOTCH_graphMapCompute(SCOTCH_Graph *, SCOTCH_Mapping *, SCOTCH_Strat *)
    int SCOTCH_graphMap(SCOTCH_Graph *, SCOTCH_Arch *, SCOTCH_Strat *, SCOTCH_Num *)
    int SCOTCH_graphPart(SCOTCH_Graph *, SCOTCH_Num, SCOTCH_Strat *, SCOTCH_Num *)

    int SCOTCH_graphOrderInit(SCOTCH_Graph *, SCOTCH_Ordering *, SCOTCH_Num *, SCOTCH_Num *, SCOTCH_Num *, SCOTCH_Num *, SCOTCH_Num *)
    void SCOTCH_graphOrderExit(SCOTCH_Graph *, SCOTCH_Ordering *)
    int SCOTCH_graphOrderLoad(SCOTCH_Graph *, SCOTCH_Ordering *, FILE *)
    int SCOTCH_graphOrderSave(SCOTCH_Graph *, SCOTCH_Ordering *, FILE *)
    int SCOTCH_graphOrderSaveMap(SCOTCH_Graph *, SCOTCH_Ordering *, FILE *)
    int SCOTCH_graphOrderSaveTree(SCOTCH_Graph *, SCOTCH_Ordering *, FILE *)
    int SCOTCH_graphOrderCompute(SCOTCH_Graph *, SCOTCH_Ordering *, SCOTCH_Strat *)
    int SCOTCH_graphOrderComputeList(SCOTCH_Graph *, SCOTCH_Ordering *, SCOTCH_Num, SCOTCH_Num *, SCOTCH_Strat *)
    int SCOTCH_graphOrderFactor(SCOTCH_Graph *, SCOTCH_Ordering *, SCOTCH_Graph *)
    int SCOTCH_graphOrderView(SCOTCH_Graph *, SCOTCH_Ordering *, FILE *)
    int SCOTCH_graphOrder(SCOTCH_Graph *, SCOTCH_Strat *, SCOTCH_Num *, SCOTCH_Num *, SCOTCH_Num *, SCOTCH_Num *, SCOTCH_Num *)
    int SCOTCH_graphOrderList(SCOTCH_Graph *, SCOTCH_Num, SCOTCH_Num *, SCOTCH_Strat *, SCOTCH_Num *, SCOTCH_Num *, SCOTCH_Num *, SCOTCH_Num *, SCOTCH_Num *)
    int SCOTCH_graphOrderCheck(SCOTCH_Graph *, SCOTCH_Ordering *)

    int SCOTCH_meshInit(SCOTCH_Mesh *)
    void SCOTCH_meshExit(SCOTCH_Mesh *)
    int SCOTCH_meshLoad(SCOTCH_Mesh *, FILE *, SCOTCH_Num)
    int SCOTCH_meshSave(SCOTCH_Mesh *, FILE *)
    int SCOTCH_meshBuild(SCOTCH_Mesh *, SCOTCH_Num, SCOTCH_Num, SCOTCH_Num, SCOTCH_Num, SCOTCH_Num *, SCOTCH_Num *, SCOTCH_Num *, SCOTCH_Num *, SCOTCH_Num *, SCOTCH_Num, SCOTCH_Num *)
    int SCOTCH_meshCheck(SCOTCH_Mesh *)
    void SCOTCH_meshSize(SCOTCH_Mesh *, SCOTCH_Num *, SCOTCH_Num *, SCOTCH_Num *)
    void SCOTCH_meshData(SCOTCH_Mesh *, SCOTCH_Num *, SCOTCH_Num *, SCOTCH_Num *, SCOTCH_Num *, SCOTCH_Num **, SCOTCH_Num **, SCOTCH_Num **, SCOTCH_Num **, SCOTCH_Num **, SCOTCH_Num *, SCOTCH_Num **, SCOTCH_Num *)
    void SCOTCH_meshStat(SCOTCH_Mesh *, SCOTCH_Num *, SCOTCH_Num *, SCOTCH_Num *, double *, double *, SCOTCH_Num *, SCOTCH_Num *, double *, double *, SCOTCH_Num *, SCOTCH_Num *, double *, double *)
    int SCOTCH_meshGraph(SCOTCH_Mesh *, SCOTCH_Graph *)
    int SCOTCH_meshGeomLoadHabo(SCOTCH_Mesh *, SCOTCH_Geom *, FILE *, FILE *, char *)
    int SCOTCH_meshGeomLoadScot(SCOTCH_Mesh *, SCOTCH_Geom *, FILE *, FILE *, char *)
    int SCOTCH_meshGeomSaveScot(SCOTCH_Mesh *, SCOTCH_Geom *, FILE *, FILE *, char *)

    int SCOTCH_meshOrderInit(SCOTCH_Mesh *, SCOTCH_Ordering *, SCOTCH_Num *, SCOTCH_Num *, SCOTCH_Num *, SCOTCH_Num *, SCOTCH_Num *)
    void SCOTCH_meshOrderExit(SCOTCH_Mesh *, SCOTCH_Ordering *)
    int SCOTCH_meshOrderSave(SCOTCH_Mesh *, SCOTCH_Ordering *, FILE *)
    int SCOTCH_meshOrderSaveMap(SCOTCH_Mesh *, SCOTCH_Ordering *, FILE *)
    int SCOTCH_meshOrderSaveTree(SCOTCH_Mesh *, SCOTCH_Ordering *, FILE *)
    int SCOTCH_meshOrderCompute(SCOTCH_Mesh *, SCOTCH_Ordering *, SCOTCH_Strat *)
    int SCOTCH_meshOrderComputeList(SCOTCH_Mesh *, SCOTCH_Ordering *, SCOTCH_Num, SCOTCH_Num *, SCOTCH_Strat *)
    int SCOTCH_meshOrder(SCOTCH_Mesh *, SCOTCH_Strat *, SCOTCH_Num *, SCOTCH_Num *, SCOTCH_Num *, SCOTCH_Num *, SCOTCH_Num *)
    int SCOTCH_meshOrderList(SCOTCH_Mesh *, SCOTCH_Num, SCOTCH_Num *, SCOTCH_Strat *, SCOTCH_Num *, SCOTCH_Num *, SCOTCH_Num *, SCOTCH_Num *, SCOTCH_Num *)
    int SCOTCH_meshOrderCheck(SCOTCH_Mesh *, SCOTCH_Ordering *)

    void SCOTCH_randomReset()

    int SCOTCH_stratInit(SCOTCH_Strat *)
    void SCOTCH_stratExit(SCOTCH_Strat *)
    void SCOTCH_stratFree(SCOTCH_Strat *)
    int SCOTCH_stratSave(SCOTCH_Strat *, FILE *)

    int SCOTCH_stratGraphBipart(SCOTCH_Strat *, char *)
    int SCOTCH_stratGraphMap(SCOTCH_Strat *, char *)
    int SCOTCH_stratGraphMapBuild(SCOTCH_Strat *, SCOTCH_Num, SCOTCH_Num, double)
    int SCOTCH_stratGraphOrder(SCOTCH_Strat *, char *)
    int SCOTCH_stratGraphOrderBuild(SCOTCH_Strat *, SCOTCH_Num, double)
    int SCOTCH_stratMeshOrder(SCOTCH_Strat *, char *)
    int SCOTCH_stratMeshOrderBuild(SCOTCH_Strat *, SCOTCH_Num, double)

    void SCOTCH_memoryTrace()
    void SCOTCH_memoryUntrace()
    void SCOTCH_memoryTraceReset()
    unsigned long SCOTCH_memoryTraceGet()



from .graph.defs cimport gint
from .graph.defs import gint_dtype


cdef gint _bisect(gint[:] a, gint x)

cdef int _is_herm_conj(complex[:, :] a, complex[:, :] b,
                       double atol=*, double rtol=*) except -1

cdef int _check_onsite(complex[:, :] M, gint norbs,
                       int check_hermiticity) except -1

cdef int _check_ham(complex[:, :] H, ham, args, params,
                    gint a, gint a_norbs, gint b, gint b_norbs,
                    int check_hermiticity) except -1

cdef void _get_orbs(gint[:, :] site_ranges, gint site,
                    gint *start_orb, gint *norbs)


cdef class BlockSparseMatrix:
    cdef public gint[:, :] block_offsets, block_shapes
    cdef public gint[:] data_offsets
    cdef public complex[:] data

    cdef complex* get(self, gint block_idx)


cdef class _LocalOperator:
    cdef public int check_hermiticity, sum
    cdef public object syst, onsite, _onsite_param_names
    cdef public gint[:, :]  where, _site_ranges
    cdef public BlockSparseMatrix _bound_onsite, _bound_hamiltonian

    cdef BlockSparseMatrix _eval_onsites(self, args, params)
    cdef BlockSparseMatrix _eval_hamiltonian(self, args, params)


cdef class Density(_LocalOperator):
    pass


cdef class Current(_LocalOperator):
    pass


cdef class Source(_LocalOperator):
    pass



# Copyright 2011-2017 Kwant authors.
#
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

import pytest
import tinyarray as ta

from sympy.physics.matrices import msigma
from sympy.physics.quantum import TensorProduct
import sympy

from kwant.continuum._common import position_operators, momentum_operators
from kwant.continuum._common import make_commutative, sympify
from kwant.continuum._common import monomials
from kwant.continuum._common import lambdify


com_A, com_B, com_C = sympy.symbols('A B C')
fA, fB, fC = sympy.symbols('A B C', cls=sympy.Function)
x_op, y_op, z_op = position_operators
kx, ky, kz = momentum_operators


@pytest.mark.parametrize('input_expr, output_expr', [
    ('k_x * A(x) * k_x', kx * fA(x_op) * kx),
    ('[[k_x * A(x) * k_x]]', sympy.Matrix([kx * fA(x_op) * kx])),
    ('k_x * sigma_y + k_y * sigma_x', kx * msigma(2) + ky * msigma(1)),
    ('[[k_x*A(x)*k_x, B(x, y)*k_x], [k_x*B(x, y), C*k_y**2]]',
     sympy.Matrix([[kx*fA(x_op)*kx, fB(x_op, y_op)*kx],
                   [kx*fB(x_op, y_op), com_C*ky**2]])),
    ('kron(sigma_x, sigma_y)', TensorProduct(msigma(1), msigma(2))),
    ('identity(2)', sympy.eye(2)),
    ('eye(2)', sympy.eye(2)),
    ('1 * sigma_x + 2 * sigma_y + 3 * sigma_z',
     msigma(1) + 2 * msigma(2) + 3 * msigma(3))
])
def test_sympify(input_expr, output_expr):
    assert sympify(input_expr) == output_expr
    assert sympify(sympify(input_expr)) == output_expr


@pytest.mark.parametrize('input_expr, output_expr, subs', [
    ('k_x', kx + ky, {'k_x': 'k_x + k_y'}),
    ('x', x_op + y_op, {'x': 'x + y'}),
    ('A', com_A + com_B, {'A': 'A + B'}),
    ('A', com_A + fB(x_op), {'A': 'A + B(x)'}),
    ('A', msigma(2), {'A': "[[0, -1j], [1j, 0]]"}),
])
def test_sympify_substitutions(input_expr, output_expr, subs):
    assert sympify(input_expr, locals=subs) == output_expr

    subs = {k: sympify(v) for k, v in subs.items()}
    assert sympify(input_expr, locals=subs) == output_expr


@pytest.mark.parametrize('subs', [
    {1: 2},
    {'1': 2},
    {'for': 33},
    {'x': 1, '1j': 7}
])
def test_sympify_invalid_substitutions(subs):
    with pytest.raises(ValueError):
        sympify('x + y', locals=subs)


@pytest.mark.parametrize('input_expr, output_expr, subs', [
    ('A + k_x**2 * eye(2)',
        kx**2 * sympy.eye(2) + msigma(2),
        {'A': "[[0, -1j], [1j, 0]]"}),
    ('A + k_x**2 * identity(2)',
        kx**2 * sympy.eye(2) + msigma(2),
        {'A': "[[0, -1j], [1j, 0]]"})

])
def test_sympify_mix_symbol_and_matrx(input_expr, output_expr, subs):
    assert sympify(input_expr, locals=subs) == output_expr

    subs = {k: sympify(v) for k, v in subs.items()}
    assert sympify(input_expr, locals=subs) == output_expr


A, B, x = sympy.symbols('A B x', commutative=False)
com_x, com_y = sympy.symbols('x y')

expr1 = x*A*x + x**2 * A * x + B*x**2

matr_com = sympy.Matrix([[expr1, expr1+A*x], [0, -expr1]])
res_mat = sympy.Matrix([[com_x**3*A + com_x**2*A + com_x**2*B, com_x**3*A + com_x**2*A + com_x**2*B + com_x*A],
                        [0, -com_x**3*A - com_x**2*A - com_x**2*B]])


def test_make_commutative():
    assert make_commutative(expr1, com_x).expand() == (make_commutative(expr1, x)).expand()
    assert make_commutative(expr1, com_x).expand() == (com_x**3*A + com_x**2*A + com_x**2*B).expand()
    assert make_commutative(matr_com, com_x).expand() == (res_mat).expand()


matr_monomials = sympify("[[x+y, a*x**2 + b*y], [y, x]]")
x, y, z = position_operators
a, b = sympy.symbols('a, b')
fA, fB = sympy.symbols('A B', cls=sympy.Function)

@pytest.mark.parametrize('expr, gens, output', [
    (x * fA(x) * x + x**2 * a, None, {x**2: fA(x), a*x**2: 1}),
    (x * fA(x) * x + x**2 * a, [x], {x**2: fA(x) + a}),
    (x**2, [x], {x**2: 1}),
    (2 * x + 3 * x**2, [x], {x: 2, x**2: 3}),
    (2 * x + 3 * x**2, 'x', {x: 2, x**2: 3}),
    (a * x**2 + 2 * b * x**2, 'x', {x**2: a + 2 * b}),
    (x**2 * (a + 2 * b) , 'x', {x**2: a + 2 * b}),
    (2 * x * y  + 3 * y * x, 'xy', {x*y: 2, y*x: 3}),
    (2 * x * a + 3 * b, 'ab', {a: 2*x, b: 3}),
    (matr_monomials, None, {
        x: sympy.Matrix([[1, 0], [0, 1]]),
        b*y: sympy.Matrix([[0, 1], [0, 0]]),
        a*x**2: sympy.Matrix([[0, 1], [0, 0]]),
        y: sympy.Matrix([[1, 0], [1, 0]])
    }),
    (matr_monomials, [x], {
        x: sympy.Matrix([[1, 0], [0, 1]]),
        1: sympy.Matrix([[y, b*y], [y, 0]]),
        x**2: sympy.Matrix([[0, a], [0, 0]])
    }),
    (matr_monomials, [x, y], {
        x: sympy.Matrix([[1, 0], [0, 1]]),
        x**2: sympy.Matrix([[0, a], [0, 0]]),
        y: sympy.Matrix([[1, b], [1, 0]])
    }),
])
def test_monomials(expr, gens, output):
    assert monomials(expr, gens) == output



@pytest.mark.parametrize("e, should_be, kwargs", [
    ("x+y", lambda x, y: x+y, dict(x=1, y=2)),
    ("1", lambda: 1, dict()),
    ("f(x)", lambda f, x: f(x), dict(f=lambda x: x, x=2)),
    (sympify("f(x)"), lambda f, x: f(x), dict(f=lambda x: x, x=2)),
    ("[[f(x)]]", lambda f, x: ta.array(f(x)), dict(f=lambda x: x, x=2)),
])
def test_lambdify(e, should_be, kwargs):
    e = lambdify(e)
    assert e(**kwargs) == should_be(**kwargs)


@pytest.mark.parametrize("e, kwargs", [
    ("x + y", dict(x=1, y=3, z=5)),
    (sympify("x+y+z"), dict(x=1, y=3, z=5)),
])
def test_lambdify_substitutions(e, kwargs):
    should_be = lambda x, y, z: x + y + z
    subs = {'y': 'y + z'}

    if not isinstance(e, str):
        # warns that 'locals' are not used if 'e' is
        # a sympy expression already
        with pytest.warns(RuntimeWarning):
            e = lambdify(e, locals=subs)
    else:
        e = lambdify(e, locals=subs)

    assert e(**kwargs) == should_be(**kwargs)



# Copyright 2011-2013 Kwant authors.
#
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

try:
    from kwant.linalg.mumps import MUMPSContext, schur_complement
    no_mumps = False
except ImportError:
    no_mumps = True

from kwant.lattice import honeycomb
from kwant.builder import Builder, HoppingKind
import pytest
import numpy as np
import scipy.sparse as sp
from ._test_utils import _Random, assert_array_almost_equal


pytestmark = pytest.mark.skipif(no_mumps, reason="MUMPS not installed")


def test_lu_with_dense():
    def _test_lu_with_dense(dtype):
        rand = _Random()
        a = rand.randmat(5, 5, dtype)
        bmat = rand.randmat(5, 5, dtype)
        bvec = rand.randvec(5, dtype)

        ctx = MUMPSContext()
        ctx.factor(sp.coo_matrix(a))

        xvec = ctx.solve(bvec)
        xmat = ctx.solve(bmat)

        assert_array_almost_equal(dtype, np.dot(a, xmat), bmat)
        assert_array_almost_equal(dtype, np.dot(a, xvec), bvec)

        # now "sparse" right hand side

        xvec = ctx.solve(sp.csc_matrix(bvec.reshape(5,1)))
        xmat = ctx.solve(sp.csc_matrix(bmat))

        assert_array_almost_equal(dtype, np.dot(a, xmat), bmat)
        assert_array_almost_equal(dtype, np.dot(a, xvec),
                                  bvec.reshape(5,1))

    _test_lu_with_dense(np.complex128)


def test_schur_complement_with_dense():
    def _test_schur_complement_with_dense(dtype):
        rand = _Random()
        a = rand.randmat(10, 10, dtype)
        s = schur_complement(sp.coo_matrix(a), list(range(3)))
        assert_array_almost_equal(dtype, np.linalg.inv(s),
                                  np.linalg.inv(a)[:3, :3])

    _test_schur_complement_with_dense(np.complex128)


def test_error_minus_9(r=10):
    """Test if MUMPSError -9 is properly caught by increasing memory"""

    graphene = honeycomb()
    a, b = graphene.sublattices

    def circle(pos):
        x, y = pos
        return x**2 + y**2 < r**2

    syst = Builder()
    syst[graphene.shape(circle, (0,0))] = -0.0001
    for kind in [((0, 0), b, a), ((0, 1), b, a), ((-1, 1), b, a)]:
        syst[HoppingKind(*kind)] = - 1

    ham = syst.finalized().hamiltonian_submatrix(sparse=True)

    # No need to check result, it's enough if no exception is raised
    MUMPSContext().factor(ham)


def test_factor_warning():
    """Test that a warning is raised if factor is asked without analysis."""
    a = sp.identity(10, dtype=complex)
    with pytest.warns(RuntimeWarning):
        MUMPSContext().factor(a, reuse_analysis=True)
